<?php

namespace Drupal\ip_geoloc\Plugin\Block;

use Drupal\Core\Block\BlockBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\ip_geoloc\Services\IpGeoLocAPI;

/**
 * IPGV&M: Set my location .
 *
 * @Block(
 *   id = "current_visitor_map",
 *   admin_label = @Translation("IPGV&M: Map (Google) centered on visitor's location"),
 *   category = @Translation("IP Geoloc"),
 * )  implements ContainerFactoryPluginInterface
 */
class CurrentVisitorMapBlock extends BlockBase implements ContainerFactoryPluginInterface {

  protected $api;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, IpGeoLocAPI $api) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->api = $api;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('ip_geoloc.api')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    // Calling the set location form in block
    // TODO find a way to add subject
    // $block['subject'] = $this->t('Your location');.
    $config = $this->getConfiguration();

    $map_options = isset($config['ip_geoloc_current_visitor_map_options']) ? $config['ip_geoloc_current_visitor_map_options'] : IP_GEOLOC_CURRENT_VISITOR_MAP_OPTIONS;

    $api = \Drupal::service('ip_geoloc.api');
    $location = $api->getVisitorLocation();

    $map_style = $config['ip_geoloc_current_visitor_map_div_style'] ? $config['ip_geoloc_current_visitor_map_div_style'] : IP_GEOLOC_MAP_DIV_DEFAULT_STYLE;

    $div_id = $config['ip_geoloc_current_visitor_map_div_id'] ? $config['ip_geoloc_current_visitor_map_div_id'] : 'ip-geoloc-block-current-visitor-map';

    // TODO check if it is correct.
    $latitude = isset($location['latitude']) ? (float) $location['latitude'] : (empty($location['fixed_address']) ? NULL : 0.0);
    $longitude = isset($location['longitude']) ? (float) $location['longitude'] : (empty($location['fixed_address']) ? NULL : 0.0);
    $info_text = isset($location['formatted_address']) ? $location['formatted_address'] : NULL;

    $settings['ip_geoloc_block'][$div_id] = [
      'ip_geoloc_current_location_map_div' => $div_id,
      'ip_geoloc_current_location_map_style' => $map_style,
      'ip_geoloc_current_location_map_options' => Json::decode($map_options),
      'ip_geoloc_current_location_map_latlng' => [$latitude, $longitude],
      'ip_geoloc_current_location_map_info_text' => $info_text,
    ];

    $build = [
      '#theme' => 'ip_geoloc_map_current_visitor',
      '#ip_geoloc_settings' => $settings['ip_geoloc_block'][$div_id],
      // '#ipgeolocsettings' => 'test',.
      '#attached' => [
        'library' => ['ip_geoloc/current_location_js'],
        'drupalSettings' => $settings,
      ],
    ];

    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheMaxAge() {
    return 0;
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state) {
    $form = parent::blockForm($form, $form_state);

    // Retrieve existing configuration for this block.
    $config = $this->getConfiguration();

    // Add fields for block configuration.
    $form['ip_geoloc_current_visitor_map_div_id'] = [
      '#type' => 'textfield',
      '#title' => t('Map div CSS id'),
      '#default_value' => $config['ip_geoloc_current_visitor_map_div_id'] ? $config['ip_geoloc_current_visitor_map_div_id'] : 'ip-geoloc-block-current-visitor-map',
      '#description' => $this->t('The map container div id to allow multiple instances of the block'),
    ];

    $form['ip_geoloc_current_visitor_map_div_style'] = [
      '#type' => 'textfield',
      '#title' => t('CSS style attribute(s) applied to the HTML DIV element that is placeholding the map'),
      '#default_value' => $config['ip_geoloc_current_visitor_map_div_style'] ? $config['ip_geoloc_current_visitor_map_div_style'] : IP_GEOLOC_MAP_DIV_DEFAULT_STYLE,
      '#description' => $this->t('The default %default_style results in a map of 300 pixels high, with a width bounded by the element that contains it. Do not enter quotes or equal signs.',
        ['%default_style' => IP_GEOLOC_MAP_DIV_DEFAULT_STYLE]),
    ];

    $form['ip_geoloc_current_visitor_map_options'] = [
      '#type' => 'textfield',
      '#size' => 120,
      '#title' => $this->t('Map options'),
      '#default_value' => $config['ip_geoloc_current_visitor_map_options'] ? $config['ip_geoloc_current_visitor_map_options'] : IP_GEOLOC_CURRENT_VISITOR_MAP_OPTIONS,
      '#description' => $this->t("The default %default_options produces a street map zoomed in to level 15. A list of map options can be found <a href='!google_map_docs'>here</a>. Remember to separate options with comma's, not semi-colons, and make sure your quotes match.",
          [
            '%default_options' => IP_GEOLOC_CURRENT_VISITOR_MAP_OPTIONS,
            '!google_map_docs' => IP_GEOLOC_DOC_GOOGLE_MAP_OPTIONS,
          ]),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    // Migration comment:  Part of ip_geoloc_block_save for current_visitor_map delta
    // Save options common to all maps.
    $this->configuration['ip_geoloc_current_visitor_map_div_style'] = $form_state->getValue('ip_geoloc_current_visitor_map_div_style');
    $this->configuration['ip_geoloc_current_visitor_map_div_id'] = $form_state->getValue('ip_geoloc_current_visitor_map_div_id');

    // @TODO Review function
    ip_geoloc_save_map_options('ip_geoloc_current_visitor_map_options', $form_state->getValue('ip_geoloc_current_visitor_map_options'));
  }

}
